setClass(Class="citizenair_stations",
         representation=representation(phenomena = "data.frame", 
                                       stations = "data.frame"))
setClass(Class="citizenair_userdata",
         representation=representation(filename = "character", 
                                       file = "character", 
                                       time = "POSIXct", 
                                       content = "list", 
                                       phenomena = "character"))

citizenair_userdata_empty <- function(){
  result <- list()
  result$filename <- NA_character_
  result$file <- NA_character_
  result$time <- Sys.time()
  result$content <- list()
  result$phenomena <- character()
  class(result) <- "citizenair_userdata"
  result
}

#' @title CitizenAir app data
#' @description CitizenAir app data
#' @docType class
#' @field endpoint object of class \code{Endpoint} as returned by \code{as.Endpoint}
#' @field stations object of class \code{citizenair_stations} as returned by \code{\link{read_stations}}
#' @field data object of class \code{citizenair_userdata} as returned by \code{\link{read_citizenair}}
#' @format The CitizenAir object is an R6Class
#' @section Methods:
#' \describe{
#'   \item{\code{fetch_timeseries(id = "123", time_span)}}{Get timeseries for Station with \code{id} and a certain time_span The argument \code{timespan.} Passed on to \code{sensorweb4R::getData}}
#'   \item{\code{getStations()}}{Get a SpatialPointsDataFrame of all stations}
#' }
#' @export
#' @examples
#' library(sensorweb4R)
#' e <- as.Endpoint("http://geo.irceline.be/sos/api/v1")
#' ca <- CitizenAir$new(e)
#' ca
#' ca$getPhenomena()
#' 
#' library(sp)
#' x <- ca$getStations()
#' plot(x)
#' 
#' library(lubridate)
#' x <- ca$fetch_timeseries(id = "1030")
#' period <- as.interval(as.POSIXct(Sys.Date() - 14), Sys.time())
#' x <- ca$fetch_timeseries(id = "1030", time_span = period)
CitizenAir <- R6::R6Class("CitizenAir",
                      public = list(
                        endpoint = NULL,
                        stations = NULL, 
                        data = NULL,
                        initialize = function(endpoint) {
                          self$endpoint <- endpoint
                          self$stations <- read_stations(endpoint)
                          self$data <- citizenair_userdata_empty()
                        },
                        fetch_timeseries = function(id, time_span=NULL) {  
                          stopifnot(length(id) == 1)
                          ## Find and create a stations object
                          x <- self$stations$stations
                          x <- as.data.frame(x)
                          x <- x[x$id %in% id, ]
                          x <- sp::SpatialPointsDataFrame(coords = x[, c("lon", "lat")], 
                                                          data = x, 
                                                          proj4string = CRS("+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0"))  
                          st <- sensorweb4R::Station(id = x$id, label = x$label, geometry = x, endpoint = self$endpoint)
                          ## Get all metadata of time series of that station
                          ts <- sensorweb4R::timeseries(self$endpoint, station = st)
                          ts <- sensorweb4R::as.Timeseries(ts)
                          ts <- sensorweb4R::fetch(ts)
                          ## Fetch the data of the specified period
                          result <- list()
                          for(i in seq_along(ts)){
                            x <- ts[i]
                            if(is.null(time_span)){
                              to <- sensorweb4R::time(sensorweb4R::lastValue(x))
                              from <- as.POSIXct(seq.Date(from = as.Date(to), by = "-1 year", length.out = 2)[2], tz = "UTC")
                              timespan <- lubridate::as.interval(from, to)  
                            }else{
                              timespan <- time_span
                            }
                            df <- sensorweb4R::getData(x, timespan = timespan)
                            df <- as.data.frame(df)
                            df$date <- as.Date(df$time)
                            df$phenomena_id <- sensorweb4R::id(sensorweb4R::Phenomenon(x))
                            df$phenomena <- self$phenomena_label(df$phenomena_id)
                            df$phenomena_label <- sensorweb4R::label(x)
                            result[[i]] <- df[, c("phenomena_id", "phenomena", "phenomena_label", "time", "value")]
                          }
                          result <- rbindlist(result)
                          result
                        },
                        phenomena_label = function(id){
                          txt_recode(id, from = self$stations$phenomena$id, to = self$stations$phenomena$label)
                        },
                        getStations = function(){
                          x <- as.data.frame(self$stations$stations)
                          x <- sp::SpatialPointsDataFrame(coords = x[, c("lon", "lat")], 
                                                          data = x, proj4string = CRS("+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0"))
                          x
                        },
                        getPhenomena = function(){
                          self$stations$phenomena$id
                        }
                      ),
                      private = list()
)
